/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_GUI_ABSTRACT_HEADER_HH_
# define WSCOUT_GUI_ABSTRACT_HEADER_HH_

# include <cstdlib>
# include <utility>

# include <wipal/pcap/pkt.hh>

# include "wscout_gui_trace.hh"
# include "wscout_gui_header_frame.hh"
# include "wscout_protocol_ids.hh"

namespace wscout
{

  namespace gui
  {

    /**
     * Interface for %protocol header widgets.
     *
     * WScout models packets as a list of %protocol headers.  Each
     * header is a Qt Widget that gets embedded inside the packet
     * widget (Packet).  In order to have Packet be independent of
     * protocols, each %protocol must satisfy a unified interface.
     * AbstractHeader specify this interface.
     *
     * To implement a new %protocol, subclass HeaderBase and implement
     * setup().
     *
     * @see HeaderBase, Packet, HeaderFrame
     */
    struct AbstractHeader
    {
      /**
       * Identify this header's next header.
       *
       * The first element specifies the next header's %protocol.  The
       * second one points to the next header's beginning of data.
       *
       * If there is no next %protocol, use protocol::NONE.  If the
       * next %protocol is unknown, use protocol::UNKNOWN.
       *
       * @see protocol::id
       */
      typedef std::pair<protocol::id, const void*>	next_header;

      /// Default destructor.
      virtual	~AbstractHeader();

      /// Frame descriptor type used by setup().
      typedef wpl::pcap::indexed_file<>::packet		packet;

      /**
       * Setup the widget for a specific packet.
       *
       * This method is <b>purely abstract</b>.  You @b must define
       * it when subclassing.
       *
       * @param[in] packet	Pointer to the packet's begining.
       * @param[in] caplen	Number of availabe bytes for this packet.
       * @param[in] trace	Packet's corresponding gui::Trace.
       * @param[in] desc	Packet's corresponding frame descriptor.
       *
       * @return The next header's type and address.
       *
       * @see next_header
       */
      virtual next_header	setup(const void*	bytes,
				      size_t		caplen,
				      const Trace&	trace,
				      const packet&	desc) = 0;
    };

  } // End of namespace gui.

} // End of namespace wscout.

#endif // ! WSCOUT_GUI_ABSTRACT_HEADER_HH_
