/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_GUI_FIND_DIALOG_HH_
# define WSCOUT_GUI_FIND_DIALOG_HH_

# include <QtGui/QStyle>
# include <QtGui/QLineEdit>
# include <QtGui/QSpinBox>
# include <QtGui/QPushButton>
# include <QtGui/QProgressBar>
# include <QtCore/QThread>

# include <wipal/pcap/stream.hh>

# include "wscout_gui_dialog.hh"

namespace wscout
{

  namespace gui
  {

    /// Thread that effectively perform searching a trace.
    class FindThread: public QThread
    {
      Q_OBJECT

    public:
      typedef wpl::pcap::indexed_file<>::const_iterator	const_iterator;

      enum
	{
	  /**
	   * Emit @c positionChanged() each time @c run() process @c
	   * positionUpdateInterval packets.
	   */
	  positionUpdateInterval	= 4096
	};

      /// Construct the thread.
      FindThread(QObject*		parent,
		 const_iterator&	current_it,
		 const const_iterator&	last_it,
		 const std::string&	bytes,
		 size_t			first,
		 size_t			last);
      /// Standard destructor.  Call cancel() if the thread is running.
      ~FindThread();

      /// Thread code.  Update @c current_it when a result is found.
      void	run();

      /// Tell whether the thread was canceled using cancel().
      bool	canceled() const;

    public slots:
      /// Tell the thread code to end prematurely and wait for termination.
      void	cancel();

    signals:
      /// Emitted each time  @c positionUpdateInterval packets are processed.
      void	positionChanged(int);

    private:
      typedef std::vector<uint8_t>	bytes_type;

      const_iterator&		current_it_;
      const const_iterator&	last_it_;

      bytes_type		bytes_;
      size_t			first_;
      size_t			last_;

      bool			canceled_;
    };

    /// Dialog box to search frames.
    class FindDialog: public Dialog
    {
      Q_OBJECT

    public:
      typedef wpl::pcap::indexed_file<>::const_iterator	const_iterator;

      /// Default constructor.
      FindDialog(QWidget* parent = 0);

      void		voidIterators();
      void		setIterators(const const_iterator&	first,
				     const const_iterator&	last,
				     size_t			last_index);


    private slots:
      void		restartClicked();
      void		findNextClicked();
      void		cancelThread();
      void		threadFinished();

    signals:
      void		found(int pos);
      void		notFound();

    private:
      bool		threadRunning() const;
      bool		validIterators() const;

      void		setupBytesRow(QGridLayout& l);
      void		setupOffsetsRow(QGridLayout& l);
      void		setupProgressBar(QGridLayout& l);
      void		setupButtonsRow(QGridLayout& l, const QStyle& s);

    private:
      typedef std::auto_ptr<FindThread>		thread_ptr;
      typedef boost::optional<const_iterator>	optional_const_iterator;

      enum
	{
	  first_min	= 0,
	  first_max	= 65535,
	  first_def	= 0,

	  last_min	= first_min,
	  last_max	= first_max,
	  last_def	= first_max
	};

      thread_ptr		thread_;
      bool			thread_canceled_;

      optional_const_iterator	first_it_;
      optional_const_iterator	current_it_;
      optional_const_iterator	last_it_;

      QLineEdit			bytes_;
      QSpinBox			first_;
      QSpinBox			last_;
      QProgressBar		pbar_;
      QPushButton		restart_;
      QPushButton		cancel_;
      QPushButton		find_next_;
      QPushButton		close_;
    };

  } // End of namespace gui.

} // End of namespace wscout.

# include "wscout_gui_find_dialog.hxx"

#endif // ! WSCOUT_GUI_FIND_DIALOG_HH_
