/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtGui/QHBoxLayout>
#include <QtGui/QScrollBar>
#include <QtGui/QApplication>
#include <QtGui/QClipboard>

#include "wscout_gui_hex_dump_dock.hh"

namespace wscout
{

  namespace gui
  {

    void
    HexDumpDock::initTextEdit(QPlainTextEdit& te, unsigned minWidth)
    {
      te.setReadOnly(true);
      te.setFont(QFont ("Monospace"));
      te.setFrameStyle(QFrame::Panel | QFrame::Sunken);
      te.setLineWrapMode(QPlainTextEdit::NoWrap);

      const QFontMetrics&	fm = te.fontMetrics();
      const int			lw = te.lineWidth();
      const int			bw = te.verticalScrollBar()->sizeHint().width();
      const int			charW = fm.width('X');
      int			mLeft, mTop, mRight, mBottom;

      te.getContentsMargins(&mLeft, &mTop, &mRight, &mBottom);

      // Add 2 to avoid having the horizontal scrollbar appearing.
      te.setMinimumWidth(minWidth * charW + lw * 2 + mLeft + mRight + bw + 2);
    }

    inline
    void
    HexDumpDock::init()
    {
      initTextEdit(addresses_, addr_len * 2);
      initTextEdit(hexData_,   bwidth * 2 + (bwidth - 1) * smallsep.size());
      initTextEdit(asciiData_, lwidth);

      addresses_.setTextInteractionFlags(Qt::NoTextInteraction);

      addresses_.setMaximumWidth(addresses_.minimumWidth());
      asciiData_.setMaximumWidth(asciiData_.minimumWidth());

      connect(addresses_.verticalScrollBar(),	SIGNAL(valueChanged(int)),
	      hexData_.verticalScrollBar(),	SLOT(setValue(int)));
      connect(addresses_.verticalScrollBar(),	SIGNAL(valueChanged(int)),
	      asciiData_.verticalScrollBar(),	SLOT(setValue(int)));
      connect(hexData_.verticalScrollBar(),	SIGNAL(valueChanged(int)),
	      addresses_.verticalScrollBar(),	SLOT(setValue(int)));
      connect(hexData_.verticalScrollBar(),	SIGNAL(valueChanged(int)),
	      asciiData_.verticalScrollBar(),	SLOT(setValue(int)));
      connect(asciiData_.verticalScrollBar(),	SIGNAL(valueChanged(int)),
	      addresses_.verticalScrollBar(),	SLOT(setValue(int)));
      connect(asciiData_.verticalScrollBar(),	SIGNAL(valueChanged(int)),
	      hexData_.verticalScrollBar(),	SLOT(setValue(int)));

      connect(&hexData_, SIGNAL(cursorPositionChanged()),
	      SLOT(hexUpdateClipboard()));
      connect(&hexData_, SIGNAL(selectionChanged()),
	      SLOT(hexUpdateClipboard()));

      {
	QHBoxLayout* const l = new QHBoxLayout;

	l->setSpacing(1);
	l->addWidget(&addresses_);
	l->addWidget(&hexData_);
	l->addWidget(&asciiData_);
	dockWidget_.setLayout(l);
      }
      setWidget(&dockWidget_);
    }

    HexDumpDock::HexDumpDock(QWidget* parent):
      QDockWidget (tr(title.c_str()), parent)
    {
      init();
    }

    HexDumpDock::HexDumpDock(const HexDumpDock& rhs):
      QDockWidget (tr(title.c_str()))
    {
      init();
      addresses_.setPlainText(rhs.addresses_.toPlainText());
      hexData_.setPlainText(rhs.hexData_.toPlainText());
      asciiData_.setPlainText(rhs.asciiData_.toPlainText());
    }

    void
    HexDumpDock::hexUpdateClipboard()
    {
      const QTextCursor&	cursor = hexData_.textCursor();

      if (cursor.hasSelection())
	{
	  const QString&	selected = cursor.selectedText();


	  if (not selected.isEmpty())
	    {
	      QClipboard&	clipboard = *QApplication::clipboard();
	      const QString	copytext = "0x" + filterBlanksOut(selected);

	      clipboard.setText(copytext, QClipboard::Selection);

	      // Also set the clipboard.  This is intended for non-X11
	      // users that have no global selection.
	      clipboard.setText(copytext, QClipboard::Clipboard);
	    }
	}
    }

    const std::string	HexDumpDock::title = "Hexadecimal dump";
    const std::string	HexDumpDock::bigsep = "  ";
    const std::string	HexDumpDock::smallsep = " ";

  } // End of namespace gui.

} // End of namespace wscout.

