/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_GUI_IEEE802_11_HEADER_HXX_
# define WSCOUT_GUI_IEEE802_11_HEADER_HXX_

# include <QtGui/QLabel>
# include <boost/foreach.hpp>

# include <wipal/wifi/frame.hh>

# include "wscout_gui_ieee802_11_header.hh"

namespace wscout
{

  namespace gui
  {

    inline
    void
    IEEE80211Header::setColor(const QColor fore, const QColor back)
    {
      QPalette	p;

      p.setColor(QPalette::WindowText, fore);
      p.setColor(QPalette::Window, back);
      setPalette(p);
    }

    inline
    IEEE80211Header::ieee80211_hooks::ieee80211_hooks(IEEE80211Header& h):
      h_ (h)
    {
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::frame_hook(const void*, size_t)
    {
      h_.duration.hide();
      h_.flags.hide();
      h_.seq_ctl.hide();
      BOOST_FOREACH(QLabel& l, h_.addr)		l.hide();
      BOOST_FOREACH(QFrame& f, h_.addr_sep)	f.hide();
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::addr_hook(const void*, size_t,
						unsigned i,
						const wpl::wifi::addr& a)
    {
      h_.addr[i - 1].setText(make_string(a).c_str());
      h_.addr[i - 1].show();
      if (i - 1)
	h_.addr_sep[i - 2].show();
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::seq_ctl_hook(const void*, size_t,
						   unsigned fragnum,
						   unsigned seqnum)
    {
      h_.seq_ctl.setText(tr("%1-%2").arg(fragnum).arg(seqnum));
      h_.seq_ctl.show();
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::
    end_of_frame_headers_hook(const void*, size_t, const void* next_header)
    {
      h_.next.second = next_header;
    }

    inline
    void
    IEEE80211Header::
    ieee80211_hooks::management_hook(const wpl::wifi::mgt::header* h, size_t)
    {
      using namespace protocol;

      static const attributes	subtypes[16] = {
	{ "Assoc. request (MGT)",    black, lightblue, IEEE802_11_ASS_REQ },
	{ "Assoc. response (MGT)",   black, lightblue, IEEE802_11_ASS_RESP },
	{ "Reassoc. request (MGT)",  black, lightblue, IEEE802_11_REASS_REQ },
	{ "Reassoc. response (MGT)", black, lightblue, IEEE802_11_REASS_RESP },
	{ "Probe request (MGT)",     black, lightblue, IEEE802_11_PROBE_REQ },
	{ "Probe response (MGT)",    black, lightblue, IEEE802_11_PROBE_RESP },
	{ "Invalid (MGT:6)",	     lightblue, red,   UNKNOWN },
	{ "Invalid (MGT:7)",	     lightblue, red,   UNKNOWN },
	{ "Beacon (MGT)",	     black, lightblue, IEEE802_11_BEACON },
	{ "Ann. traffic indic. mess. (MGT)", black, lightblue, NONE },
	{ "Disassoc. (MGT)",	     black, lightblue, IEEE802_11_DISASSOC },
	{ "Auth. (MGT)",	     black, lightblue, IEEE802_11_AUTH },
	{ "Deauth. (MGT)",	     black, lightblue, IEEE802_11_DEAUTH },
	{ "Action (MGT)",	     black, lightblue, IEEE802_11_ACTION },
	{ "Invalid (MGT:14)",	     lightblue, red,   UNKNOWN },
	{ "Invalid (MGT:15)",	     lightblue, red,   UNKNOWN },
	};

      const unsigned		s = wpl::wifi::subtype_of(h);
      const bool		p = wpl::wifi::control_flag(6, h);
      const protocol::id	i = subtypes[s].next;

      h_.next.first =  i == UNKNOWN or i == NONE or not p ? i : ENCRYPTED;
      setDesc(subtypes[s]);
      setControl(h);
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::control_hook(const void* h, size_t)
    {
      static const attributes	subtypes[16] =
	{
	  { "Invalid (CTL:0)",		yellow, red,	protocol::NONE },
	  { "Invalid (CTL:1)",		yellow, red,	protocol::NONE },
	  { "Invalid (CTL:2)",		yellow, red,	protocol::NONE },
	  { "Invalid (CTL:3)",		yellow, red,	protocol::NONE},
	  { "Invalid (CTL:4)",		yellow, red,	protocol::NONE },
	  { "Invalid (CTL:5)",		yellow, red,	protocol::NONE },
	  { "Invalid (CTL:6)",		yellow, red,	protocol::NONE },
	  { "Invalid (CTL:7)",		yellow, red,	protocol::NONE },
	  { "Block Ack. Req. (CTL)",	black, yellow,	protocol::NONE },
	  { "Block Ack. (CTL)",		black, yellow,	protocol::NONE },
	  { "PS-Poll (CTL)",		black, yellow,	protocol::NONE },
	  { "RTS (CTL)",		black, yellow,	protocol::NONE },
	  { "CTS (CTL)",		black, yellow,	protocol::NONE },
	  { "Ack. (CTL)",		black, yellow,	protocol::NONE },
	  { "CF-End (CTL)",		black, yellow,	protocol::NONE },
	  { "CF-Ack (CTL)",		black, yellow,	protocol::NONE },
	};

      const unsigned		s = wpl::wifi::subtype_of(h);

      h_.next.first = subtypes[s].next;
      setDesc(subtypes[s]);
      setControl(h);
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::data_hook(const void* h, size_t)
    {
      using namespace protocol;

      static const attributes	subtypes[16] =
	{
	  { "Data",				black, lightgreen, LLC },
	  { "Data + CF-Ack",			black, lightgreen, LLC },
	  { "Data + CF-Poll",			black, lightgreen, LLC },
	  { "Data + CF-Ack + CF-Poll",		black, lightgreen, LLC },
	  { "Null (no data)",			black, lightgreen, NONE },
	  { "CF-Ack (no data)",			black, lightgreen, NONE },
	  { "CF-Poll (no data)",		black, lightgreen, NONE },
	  { "CF-Ack + CF-Poll (no data)",	black, lightgreen, NONE },
	  { "QoS Data",				black, lightgreen, LLC },
	  { "QoS Data + CF-Ack",		black, lightgreen, LLC },
	  { "QoS Data + CF-Poll",		black, lightgreen, LLC },
	  { "QoS Data + CF-Ack + CF-Poll",	black, lightgreen, LLC },
	  { "QoS Null (no data)",		black, lightgreen, NONE },
	  { "Invalid (DATA:13)",		lightgreen, red,   UNKNOWN },
	  { "QoS CF-Poll (no data)",		black, lightgreen, NONE },
	  { "QoS CF-Ack + CF-Poll (no data)",	black, lightgreen, NONE },
	};

      const unsigned		s = wpl::wifi::subtype_of(h);
      const bool		p = wpl::wifi::control_flag(6, h);
      const protocol::id	i = subtypes[s].next;

      h_.next.first =  i == UNKNOWN or i == NONE or not p ? i : ENCRYPTED;
      setDesc(subtypes[s]);
      setControl(h);
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::invalid_type_hook(const void* h, size_t)
    {
      const unsigned	type = wpl::wifi::type_of(h);
      const unsigned	subtype = wpl::wifi::subtype_of(h);

      h_.setColor(white, red);
      h_.desc.setText(tr("Invalid frame type (%1:%2)").arg(type).arg(subtype));
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::
    truncated_frame_hook(const void*, size_t,
			 wpl::wifi::dissector_status::status s)
    {
      static const attributes truncated =
	{ "Truncated frame", white, red, protocol::NONE };

      if (s == wpl::wifi::dissector_status::invalid)
	setDesc(truncated);
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::setDesc(const attributes& attr)
    {
      h_.setColor(attr.fore, attr.back);
      h_.desc.setText(tr(attr.desc.c_str()));
    }

    inline
    void
    IEEE80211Header::ieee80211_hooks::setControl(const void* f)
    {
      // Duration
      h_.duration.setText(tr("duration: %1").arg(wpl::wifi::duration_of(f)));
      h_.duration.show();

      // Flags.
      {
	static const char	set[] = "TFGRPDWO";
	char			res[sizeof (set)];

	for (size_t i = 0; i < sizeof (set) - 1; ++i)
	  res[i] = wpl::wifi::control_flag(i, f) ? set[i] : '-';
	res[sizeof (set) - 1] = '\0';
	h_.flags.setText(res);
	h_.flags.show();
      }
    }

  } // End of namespace gui.

} // End of namespace wscout.

#endif // ! WSCOUT_GUI_IEEE802_11_HEADER_HXX_
