/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtGui/QMenuBar>
#include <QtGui/QFileDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QToolBar>
#include <QtGui/QStatusBar>

#include "boost/foreach.hpp"

#include <wipal/pcap/stream.hh>
#include <wipal/tool/exceptions.hh>

#include "wscout_gui_main_window.hh"
#include "wscout_gui_filter_dialog.hh"
#include "wscout_gui_pixmap.hh"
#include "wscout_config.hh"

/// WScout constructs.
namespace wscout
{

  /// WScout's GUI elements.
  namespace gui
  {

    MainWindow::MainWindow():
      focus_finder_ (QKeySequence::Find, this),
      current_search_ (this)
    {
      initWidgets();
      updateTrace();
    }

    MainWindow::MainWindow(const MainWindow& w):
      QMainWindow (),
      focus_finder_ (QKeySequence::Find, this),
      hexFrameDock_ (w.hexFrameDock_),
      current_search_ (this)
    {
      initWidgets(&w);

      if (const int c = w.tabs_.count())
	for (int i = 0; i < c; ++i)
	  addTrace(new Trace (*dynamic_cast<const Trace*> (w.tabs_.widget(i))));
      else
	updateTrace();
    }

    bool
    MainWindow::open(const QString& directoryHint)
    {
      const QString dir =
	directoryHint.isNull() ? openDirectoryHint() : directoryHint;

      const QString traceName =
	QFileDialog::getOpenFileName(this, tr("Open trace"), dir);

      if (traceName.isNull())
	return false;
      return openTrace(traceName);
    }

    bool
    MainWindow::openTrace(const QString& traceName)
    {
      indexed_file*		pcapTrace = 0;
      QProgressDialog*		dialog = 0;
      NewProgressDialog		setup (&dialog, this);
      UpdateProgressDialog	update (&dialog);

      try
        {
          pcapTrace = new indexed_file (traceName.toStdString(),
					setup, update);
        }
      catch (const wpl::tool::user_interruption&)
	{
	  pcapTrace = 0;
	}
      catch (const wpl::tool::file_error& e)
        {
	  pcapTrace = 0;
          QMessageBox::critical(this, tr("Error"), e.what());
        }
      catch (const std::invalid_argument& e)
        {
	  pcapTrace = 0;
          QMessageBox::critical(this, tr("Error"), e.what());
        }
      delete dialog;
      if (not pcapTrace)
	return false;

      addTrace(new Trace (pcapTrace));
      return true;
    }

    bool
    MainWindow::openInNewWindow() const
    {
      MainWindow* const w = newWindow(false);

      assert(w);
      if (not w->open(openDirectoryHint()))
	{
	  delete w;
	  return false;
	}
      w->show();
      return true;
    }

    MainWindow*
    MainWindow::duplicateWindow() const
    {
      return setupNewWindow(new MainWindow (*this), true);
    }

    MainWindow*
    MainWindow::newWindow(bool show) const
    {
      return setupNewWindow(new MainWindow, show);
    }

    void
    MainWindow::closeTab()
    {
      if (1 == tabs_.count())
	close();
      trace()->deleteLater();
    }

    void
    MainWindow::filter()
    {
      Trace* const current = trace();
      assert(current);

      const QString	trace_name = (boost::lexical_cast<std::string>
				      (*current->trace()).c_str());
      const QString	default_out = trace_name + ".filtered";

      const QString traceName =
	FilterDialog::getFilteredTraceName(this,
					   trace_name,
					   tr("Filter trace"),
					   default_out);

      if (not traceName.isNull())
	openTrace(traceName);
    }

    void
    MainWindow::find(bool show)
    {
      const Trace* const		current = trace();

      if (current)
	{
	  const indexed_file* const	pcapTrace = current->trace();
	  assert(pcapTrace);

	  const QString	fileName = (boost::lexical_cast<std::string>
				    (*pcapTrace).c_str());
	  const QString	shortName = QFileInfo (fileName).fileName();

	  current_search_.setWindowTitle(tr("Find ") + shortName);
	  current_search_.setIterators(pcapTrace->begin(), pcapTrace->end(),
				       pcapTrace->size());
	  if (show)
	    current_search_.show();
	}
    }

    void
    MainWindow::stepTab(int step)
    {
      const int	count = tabs_.count();

      if (count < 2)
	return;

      // Add an extra "count" to avoid the mathematically unsound
      // behavior of C's modulo (%) operator with negative numbers.
      const int n = (tabs_.currentIndex() + count + step) % count;
      tabs_.setCurrentIndex(n);
    }

    void
    MainWindow::prevTab()
    {
      stepTab(-1);
    }

    void
    MainWindow::nextTab()
    {
      stepTab(1);
    }

    void
    MainWindow::begin()
    {
      assert(trace());

      trace()->setPosition(0);
    }

    void
    MainWindow::end()
    {
      assert(trace());

      trace()->setPosition(trace()->lastPosition());
    }

    void
    MainWindow::forward()
    {
      assert(trace());

      trace()->setPosition(trace()->position() + step_.currentText().toInt());
    }

    void
    MainWindow::backward()
    {
      assert(trace());

      trace()->setPosition(trace()->position() - step_.currentText().toInt());
    }

    void
    MainWindow::about()
    {
      // FIXME: Find a mechanism to auto-magically keep this text in sync
      // FIXME: with README.
      static const char text[] =
	"<h1 align=\"center\" style=\"margin-bottom: 3px\">%1</h1>"
	"<p align=\"center\" style=\"margin-top: 3px; margin-bottom: 22px\">"
	"  <a href=\"%5\">%5</a>"
	"</p>"
	"<p>%1 is a lightweight PCAP file visualizer.</p>"
	"<p>"
	"%1 provides a trace visualizer that is able to work with "
	"huge traces <nobr>(>10 GiB)</nobr>.  Its goals are speed and "
	"low memory requirements.  Despite its design being "
	"protocol-agnostic, it currently handles only Prism and "
	"IEEE 802.11 headers."
	"</p>"
	"<p><dl>"
	"  <dt><b>Version</b></dt>		<dd>%2</dd>"
	"  <dt><b>Report bugs to</b></dt>	<dd><a href=\"%3\">%3</a></dd>"
	"  <dt><b>Main author</b></dt>		<dd>Thomas Claveirole "
	"    &lt;<a href=\"mailto:thomas.claveirole@lip6.fr\">"
	"thomas.claveirole@lip6.fr</a>&gt;</dt>"
	"  <dt><b>Organization</b></dt>		<dd>%4</dd>"
	"</dl></p>"
	"<p style=\"margin-top: 22px;"
	"  font-style: italic\">"
	"  &copy; %4</p>"
	"<p style=\"font-style: italic\">"
	"  This is free software; see the source for copying conditions.  "
	"  There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR "
	"  A PARTICULAR PURPOSE."
	"</p>";

      QMessageBox::about(this,
			 tr("About %1").arg(qApp->applicationName()),
			 tr(text)
			 .arg(qApp->applicationName())		// %1
			 .arg(PACKAGE_VERSION)			// %2
			 .arg(PACKAGE_BUGREPORT)		// %3
			 .arg(qApp->organizationName())		// %4
			 .arg(qApp->organizationDomain()));	// %5
    }

    bool
    MainWindow::writeSelection(QString file, bool open)
    {
      assert(trace());

      if (file.isNull())
	file = QFileDialog::getSaveFileName(this, tr("Save selected packets"),
					    openDirectoryHint());
      if (file.isNull())
	return false;

      try
	{
	  trace()->writeSelection(file);
	}
      catch (const std::exception&)
	{
	  return false;
	}
      if (open)
	return openTrace(file);
      return true;
    }

    void
    MainWindow::found(int pos)
    {
      trace()->setPosition(pos);
    }

    void
    MainWindow::notFound()
    {
      QMessageBox::information(this,
			       tr("Not found"), tr("End of trace reached."));
    }

    void
    MainWindow::tabChanged(int)
    {
      find(false);
    }

    void
    MainWindow::updateTrace()
    {
      Trace* const t = trace();

      if (t)
	{
	  BOOST_FOREACH(QAction* a, trace_related_actions_)
	    a->setEnabled(true);
	  finder_.setEnabled(true);
	  finder_.setMinimum(1);
	  finder_.setMaximum(t->lastPosition() + 1);

	  swap_.setCheckState(t->swapContent() ? Qt::Checked : Qt::Unchecked);
	  swap_.setEnabled(true);

	  const indexed_file* const	pcapT = t->trace();
	  const QString			fileName =
	    boost::lexical_cast<std::string> (*pcapT).c_str();

	  trace_name_.setText(fileName);
	  updatePosition(t->position());
	  trace_size_.setText(tr("%1 MiB")
			      .arg(pcapT->filesize() / float (1 << 20)));

	  setWindowTitle(QFileInfo (fileName).fileName() + " - " +
			 qApp->applicationName());

	  t->setFocus();
	}
      else
	{
	  static const QString	na = tr("N/A");

	  BOOST_FOREACH(QAction* a, trace_related_actions_)
	    a->setEnabled(false);
	  finder_.setEnabled(false);
	  swap_.setEnabled(false);
	  trace_name_.setText(na);
	  position_.setText(na);
	  trace_size_.setText(na);
	  setWindowTitle(qApp->applicationName());
	}

      const bool tabs_actions_enabled = tabs_.count() > 1;
      BOOST_FOREACH(QAction* a, tabs_related_actions_)
	a->setEnabled(tabs_actions_enabled);
    }

    void
    MainWindow::updatePosition(int p)
    {
      assert(trace());

      position_.setText(trace()->positionString());
      finder_.setValue(p + 1);
    }

    void
    MainWindow::FinderEditingFinished()
    {
      assert(trace());

      trace()->setPosition(finder_.value() - 1);
    }

    void
    MainWindow::swapContent(int state)
    {
      assert(trace());

      switch (state)
	{
	case Qt::Unchecked:
	  trace()->setSwapContent(false);
	  break;

	case Qt::Checked:
	  trace()->setSwapContent(true);
	  break;

	default:
	  assert(! "Internal error, please report this bug.");
	  break;
	}
    }

    void
    MainWindow::updateHexDock(int p)
    {
      Trace* const		t = trace();
      assert(t);

      indexed_file* const	d = t->trace();
      assert(d);

      const packet		f = *(*d)[p];

      hexFrameDock_.setup(f.bytes(), f.meta().caplen);
    }

    void
    MainWindow::showHexDock(int p)
    {
      updateHexDock(p);
      hexFrameDock_.show();
    }

    void
    MainWindow::setupMenuBar()
    {
      // File.
      {
	QMenu* const	fileMenu = menuBar()->addMenu(tr("&File"));
	QAction* const	o = fileMenu->addAction(tr("&Open..."),
						this, SLOT(open()));
	QAction* const	n = fileMenu->addAction(tr("Open in &new window..."),
						this, SLOT(openInNewWindow()));
	QAction* const	w = fileMenu->addAction(tr("Write selected packets..."),
						this, SLOT(writeSelection()));
	QAction* const	c = fileMenu->addAction(tr("&Close"),
						this, SLOT(closeTab()));
	fileMenu->addSeparator();
	QAction* const	q = fileMenu->addAction(tr("&Quit"),
						qApp, SLOT(quit()));

	o->setShortcut(QKeySequence::Open);
	n->setShortcut(QKeySequence::New);
	w->setShortcut(QKeySequence::Save);
	c->setShortcut(QKeySequence::Close);
	q->setShortcut(tr("Ctrl+Q"));

	trace_related_actions_.push_back(w);
	trace_related_actions_.push_back(c);
      }

      // Actions.
      {
	QMenu* const	actMenu = menuBar()->addMenu(tr("&Actions"));
	QAction* const	f = actMenu->addAction(tr("&Find..."),
					       this, SLOT(find()));
	QAction* const	i = actMenu->addAction(tr("F&ilter..."),
						this, SLOT(filter()));
	actMenu->addSeparator();
	QAction* const	b = actMenu->addAction(tr("&Beginning"),
						this, SLOT(begin()));
	QAction* const	a = actMenu->addAction(tr("Step b&ackward"),
						this, SLOT(backward()));
	QAction* const	o = actMenu->addAction(tr("Step f&orward"),
						this, SLOT(forward()));
	QAction* const	e = actMenu->addAction(tr("&End"), this, SLOT(end()));

	f->setShortcut(QKeySequence::Find);
	i->setShortcut(tr("Ctrl+I"));
	b->setShortcut(QKeySequence::MoveToStartOfDocument);
	a->setShortcut(QKeySequence::Back);
	o->setShortcut(QKeySequence::Forward);
	e->setShortcut(QKeySequence::MoveToEndOfDocument);

	trace_related_actions_.push_back(f);
	trace_related_actions_.push_back(i);
	trace_related_actions_.push_back(b);
	trace_related_actions_.push_back(a);
	trace_related_actions_.push_back(o);
	trace_related_actions_.push_back(e);
      }

      // View.
      {
	QMenu* const	viewMenu = menuBar()->addMenu(tr("&View"));

	QAction* const	w = viewMenu->addAction(tr("New &window"),
						this, SLOT(newWindow()));
	QAction* const	d = viewMenu->addAction(tr("&Duplicate window"),
						this, SLOT(duplicateWindow()));
	viewMenu->addSeparator();


	QAction* const	p = viewMenu->addAction(tr("&Previous tab"),
						this, SLOT(prevTab()));
	QAction* const	n = viewMenu->addAction(tr("&Next tab"),
						this, SLOT(nextTab()));
	viewMenu->addSeparator();
	QAction* const	s = viewMenu->addAction(tr("&Swap endianness"),
						&swap_, SLOT(toggle()));
	QAction* const	h = hexFrameDock_.toggleViewAction();
	viewMenu->addAction(h);

	w->setShortcut(QKeySequence::AddTab);
	d->setShortcut(tr("Ctrl+D"));
	p->setShortcut(QKeySequence::PreviousChild);
	n->setShortcut(QKeySequence::NextChild);
	s->setShortcut(tr("Ctrl+E"));
	h->setShortcut(tr("Ctrl+H"));

	tabs_related_actions_.push_back(p);
	tabs_related_actions_.push_back(n);

	trace_related_actions_.push_back(s);
      }

      // Help.
      {
	QMenu* const	helpMenu = menuBar()->addMenu(tr("&Help"));
	helpMenu->addAction(tr("About &%1...").arg(qApp->applicationName()),
			    this, SLOT(about()));
	helpMenu->addAction(tr("About &Qt..."), qApp, SLOT(aboutQt()));
      }
    }

    void
    MainWindow::setupToolBar()
    {
      QToolBar* const t = new QToolBar;

      // File operations.
      {
	t->addAction(pixmap(tr("fileopen")),
		     tr("Open..."), this, SLOT(open()));
	QAction* const c = t->addAction(pixmap(tr("fileclose")),
					tr("Close"), this, SLOT(closeTab()));
	trace_related_actions_.push_back(c);
      }
      // View operations.
      {
	t->addSeparator();
	t->addAction(pixmap(tr("window_new")), tr("New window"),
		     this, SLOT(newWindow()));
	t->addAction(pixmap(tr("window_dup")), tr("Duplicate window"),
		     this, SLOT(duplicateWindow()));
	t->addAction(pixmap(tr("window_openfile")), tr("Open in new window..."),
		     this, SLOT(openInNewWindow()));
      }
      // Actions.
      {
	t->addSeparator();
	QAction* const f = t->addAction(pixmap(tr("filter")),
					tr("Filter..."), this, SLOT(filter()));
	QAction* const i = t->addAction(pixmap(tr("find")),
					tr("Find..."), this, SLOT(find()));
	trace_related_actions_.push_back(f);
	trace_related_actions_.push_back(i);
      }
      // Step chooser.
      {
	step_.setEditable(true);
	step_.setInsertPolicy(QComboBox::InsertAlphabetically);
	step_.setToolTip("Step amount");
	// FIXME: Memory leak!
	step_.setValidator(new QIntValidator (1, int (1e9), &step_));
	step_.addItem("1");
	step_.addItem("5");
	step_.setCurrentIndex(1);
	step_.addItem("10");
	step_.addItem("25");
	step_.addItem("50");
	step_.addItem("100");
	step_.addItem("500");
	step_.addItem("1000");
	step_.addItem("5000");
	step_.addItem("10000");
	step_.addItem("100000");
	t->addWidget(&step_);
      }

      // Navigation misc.
      {
	QAction* const b = t->addAction(pixmap(tr("player_start")),
					tr("Beginning"),
					this, SLOT(begin()));
	QAction* const a = t->addAction(pixmap(tr("player_rew")),
					tr("Step backward"),
					this, SLOT(backward()));
	QAction* const f = t->addAction(pixmap(tr("player_fwd")),
					tr("Step forward"),
					this, SLOT(forward()));
	QAction* const e = t->addAction(pixmap(tr("player_end")),
					tr("End"), this, SLOT(end()));

	trace_related_actions_.push_back(b);
	trace_related_actions_.push_back(a);
	trace_related_actions_.push_back(f);
	trace_related_actions_.push_back(e);
      }

      // Finder.
      {
	finder_.setPrefix("#");
	finder_.setSizePolicy(QSizePolicy (QSizePolicy::Expanding,
					   QSizePolicy::Fixed));
	finder_.setToolTip("Goto packet");
	connect(&finder_,
		SIGNAL(editingFinished()),
		SLOT(FinderEditingFinished()));
	connect(&focus_finder_,	SIGNAL(activated()),
		&finder_,	SLOT(clear()));
	connect(&focus_finder_,	SIGNAL(activated()),
		&finder_,	SLOT(setFocus()));
	t->addWidget(&finder_);
      }

      // Reverse endianness.
      {
	swap_.setText(tr("Reverse endianness"));
	swap_.setToolTip(tr("Reverse this trace header's endianness when "
			    "that makes sense (e.g. for Prism headers.)"));
	connect(&swap_, SIGNAL(stateChanged(int)), SLOT(swapContent(int)));
	t->addWidget(&swap_);
      }
      addToolBar(t);
    }

    void
    MainWindow::setupStatusBar()
    {
      const QSizePolicy	p (QSizePolicy::MinimumExpanding,
			   QSizePolicy::Minimum);
      QStatusBar* const s = statusBar();

      trace_name_.setSizePolicy(p);
      position_.setSizePolicy(p);
      trace_size_.setSizePolicy(p);

      trace_name_.setContentsMargins(status_bar_hmargin, status_bar_vmargin,
				     status_bar_hmargin, status_bar_vmargin);
      position_.setContentsMargins(status_bar_hmargin, status_bar_vmargin,
				   status_bar_hmargin, status_bar_vmargin);
      trace_size_.setContentsMargins(status_bar_hmargin, status_bar_vmargin,
				     status_bar_hmargin, status_bar_vmargin);

      s->addWidget(&trace_name_);
      s->addWidget(&position_);
      s->addWidget(&trace_size_);
    }

    void
    MainWindow::setupCentralWidget()
    {
      setCentralWidget(&tabs_);
      connect(&tabs_, SIGNAL(currentChanged(int)), SLOT(updateTrace()));
    }

    void
    MainWindow::setupHexDock(const MainWindow* w)
    {
      addDockWidget(Qt::BottomDockWidgetArea, &hexFrameDock_);
      hexFrameDock_.setVisible(w ? w->hexFrameDock_.isVisible() : false);
    }

    void
    MainWindow::initWidgets(const MainWindow* w)
    {
      setWindowIcon(pixmap(tr("wscout")));
      setupMenuBar();
      setupToolBar();
      setupCentralWidget();
      setupStatusBar();
      setupHexDock(w);

      connect(&tabs_, SIGNAL(currentChanged(int)), this, SLOT(tabChanged(int)));
      connect(&current_search_, SIGNAL(found(int)), this, SLOT(found(int)));
      connect(&current_search_, SIGNAL(notFound()), this, SLOT(notFound()));

      resize(950, 500);
    }


  } // End of namespace gui.

} // End of namespace wscout.
